# * @author: Jiang Tao (tjiang@hit.edu.cn)

from collections import Counter

def acquire_count_max(_list_):
	c = Counter(_list_)
	return c.most_common(1)[0]
	# this is a tuple

def construct_concensus_info(Ins_list, Clip_list, evidence_read, SV_size):
	total_count = len(Ins_list) + len(Clip_list)
	if total_count < evidence_read:
		return 0
	breakpoint = list()
	insert_size = list()
	boundary = list()
	for i in Ins_list:
		breakpoint.append(i[0])
		insert_size.append(i[1])
	for i in Clip_list:
		if i[2] == 1:
			breakpoint.append(i[0])

	# ==============method_1=====================
	Prob_pos_1 = Counter(breakpoint).most_common(1)[0][0]
	# ==============method_2=====================
	Prob_pos_2 = sum(breakpoint)/len(breakpoint)
	Average_size = int(sum(insert_size)/len(insert_size))
	if Average_size < SV_size:
		return 0

	local_info = list()
	local_name = [Prob_pos_2, Average_size]
	local_id = 0
	for i in Ins_list:
		info = local_name + [str(local_id), i[2]]
		local_id += 1
		local_info.append(info)
	for i in Clip_list:
		info = local_name + [str(local_id), i[1]]
		local_id += 1
		local_info.append(info)

	return local_info


def construct_concensus_seq(Ins_list, Clip_list):
	'''
	Ins_list: 	start position on reference genome
				Insertion size
				Insertion sequence
	Clip_list:	clip position on reference genome
				clip sequence
				clip type(0 for left and 1 for right)
	'''
	breakpoint = list()
	insert_size = list()
	for i in Ins_list:
		breakpoint.append(i[0])
		insert_size.append(i[1])
	for i in Clip_list:
		if i[2] == 1:
			breakpoint.append(i[0])

	# ==============method_1=====================
	Prob_pos_1 = Counter(breakpoint).most_common(1)[0][0]
	# ==============method_2=====================
	Prob_pos_2 = sum(breakpoint)/len(breakpoint)
	Max_size = max(insert_size)
	Min_size = min(insert_size)
	Average_size = int(sum(insert_size)/len(insert_size))

	Seq = dict()
	for i in Ins_list:
		for j in xrange(i[1]):
			pos = i[0] + j
			ch = i[2][j]
			if pos not in Seq:
				Seq[pos] = list()
			Seq[pos].append(ch)

	for i in Clip_list:
		if Average_size <= len(i[1]):
			boundary = Average_size
		else:
			boundary = len(i[1])

		if i[2] == 0:
			local_clip_seq = i[1][len(i[1])-boundary:]
			for j in xrange(boundary):
				pos = i[0] + j
				ch = local_clip_seq[j]
				if pos not in Seq:
					Seq[pos] = list()
				Seq[pos].append(ch)
		else:
			for j in xrange(boundary):
				pos = i[0] + j
				ch = i[1][j]
				if pos not in Seq:
					Seq[pos] = list()
				Seq[pos].append(ch)
	Seq_trans = list()
	for key in Seq:
		if len(Seq[key]) < 5:
			continue
		Seq_trans.append([key, acquire_count_max(Seq[key])[0]])
	Seq_trans = sorted(Seq_trans, key = lambda x:x[0])
	final_consensus = str()
	for i in Seq_trans:
		if i[0] < Prob_pos_1:
			continue
		final_consensus += i[1]
		if len(final_consensus) > Average_size:
			break
	return final_consensus, Prob_pos_1
	


#!/usr/bin/env python

''' 
 * All rights Reserved, Designed By HIT-Bioinformatics   
 * @Title:  call_TE.py
 * @Package: argparse, sys, logging, pysam, Bio
 * @Description: Establish the ME callset
 * @author: Jiang Tao (tjiang@hit.edu.cn)
 * @date: Apr 24 2018
 * @version V1.0.2  
'''

import argparse
import logging
import sys
import time
import cigar

from collections import Counter
from rMETL.rMETL_version import __version__, __author__, __contact__
from rMETL.rMETL_genotype import simple_call_genotype
from rMETL.rMETL_cmdRunner import setupLogging
from rMETL.rMETL_utils import load_ref

USAGE="""\
           _  ___  _   _____   _______   _
     _ _  | ^_   _^ | |  ___| |__   __| | |
    | ^_| | | | | | | | |__      | |    | |
    | |   | | | | | | |  __|     | |    | |
    | |   | | | | | | | |___     | |    | |___
    |_|   |_| |_| |_| |_____|    |_|    |_____|

    rMETL - realignment-based Mobile Element insertion detection Tool for Long read

	Generate final MEI/MED callset in bed or vcf file.
	
	The output file called 'calling.bed' or 'calling.vcf'
	stores in output directory.
	
	rMETL V%s
	Author: %s
	Contact: %s
"""%(__version__, __author__, __contact__)

def acquire_count_max(_list_):
	c = Counter(_list_)
	return c.most_common(1)[0][0]

flag_dic = {0:1, \
			16:2, \
			256:0, \
			272:0, \
			2048:0, \
			2064:0, \
			4:0}

STRAND = {'1':'+', \
		  '2':'-', \
		  '*':'+-'}

cluster_dic = {}

strand_dic = {1:'+', \
			  2:'-'}

class R_INFO(object):
	"""store the infomation of the signal sequence"""
	def __init__(self, Type, Chr, Pos, Len, GT):
		self.Type = Type
		self.Chr = Chr
		self.Pos = Pos
		self.Len = Len
		self.GT = GT

def parse_name(seq):
	chr = seq.split('*')[0]
	breakpoint = seq.split('*')[1]
	insert_size = seq.split('*')[2]
	GT = seq.split('*')[3]
	return chr, breakpoint, insert_size, GT

def parse_name_tp(line):
	'''
	resolution signatures for bed format
	'''
	seq = line.split('*')
	Type = seq[0]
	chr = seq[1]
	pos = seq[2]
	len = seq[3]
	if Type == 'DEL':
		rc = seq[4]
		cov = seq[5]
	else:
		rc = seq[5]
		cov = seq[6]
	GT = rc+':'+cov
	local_info = R_INFO(Type, chr, pos, len, GT)
	return local_info

def clip_analysis(deal_cigar, clipping_threshold):
	'''
	resolution cogar
	'''
	seq = list(cigar.Cigar(deal_cigar).items())
	if seq[0][1] == 'S':
		first_pos = seq[0][0]
	else:
		first_pos = 0
	if seq[-1][1] == 'S':
		last_pos = seq[-1][0]
	else:
		last_pos = 0
	total_len = first_pos + last_pos
	signal_len = 0
	for i in seq:
		signal_len += i[0]
	if signal_len == 0:
		return 0
	if total_len*1.0 / signal_len >= clipping_threshold:
		return 0
	else:
		return 1

def print_vcf_head(ref, sample):
	'''
	generation of VCF head
	'''
	import time
	Date = time.strftime("%Y%m%d")
	head = list()
	head.append("##fileformat=VCFv4.2\n")
	head.append("##fileDate=%s\n"%(Date))
	head.append("##source=rMETL\n")
	for i in ref:
		head.append("##contig=<ID=%s,length=%d>\n"%(i, len(ref[i])))
	head.append("##ALT=<ID=<DEL>,Description=\"Deletion relative to the reference\">\n")
	head.append("##ALT=<ID=<INS>,Description=\"Insertion of sequence relative to the reference\">\n")
	head.append("##INFO=<ID=END,Number=1,Type=Integer,Description=\"End position of the variant described in this record\">\n")
	head.append("##INFO=<ID=SVLEN,Number=.,Type=String,Description=\"Difference in length between REF and ALT alleles\">\n")
	head.append("##INFO=<ID=IMPRECISE,Number=0,Type=Flag,Description=\"Imprecise structural variation\">\n")
	head.append("##INFO=<ID=PRECISE,Number=0,Type=Flag,Description=\"Precise structural variation\">\n")
	head.append("##INFO=<ID=AC,Number=.,Type=Integer,Description=\"Allele count'\">\n")
	head.append("##INFO=<ID=AF,Number=.,Type=Float,Description=\"Allele frequency'\">\n")
	head.append("##INFO=<ID=AN,Number=.,Type=String,Description=\"Allele name'\">\n")
	head.append("##FORMAT=<ID=GT,Number=1,Type=String,Description=\"Genotype\">\n")
	head.append("##FORMAT=<ID=DV,Number=1,Type=Integer,Description=\"#High-quality variant reads\">\n")
	head.append("##FORMAT=<ID=DR,Number=1,Type=Integer,Description=\"#Reference reads\">\n")
	head.append("#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\t%s\n"%(sample))
	return head

def parse_seq_head(line):
	'''
	resolution signatures for vcf format
	'''
	seq = line.split('*')
	Type = seq[0]
	chr = seq[1]
	pos = seq[2]
	len = seq[3]
	if Type == 'DEL':
		rc = seq[4]
		cov = seq[5]
	else:
		rc = seq[5]
		cov = seq[6]
	GT = rc+':'+cov
	local_info = R_INFO(Type, chr, pos, len, GT)
	return local_info

# ************************BED_FUNCTION*******************************
def call_bed(args):
	path = args.input
	out_path = args.output + "calling.bed"
	AlignmentFile = open(path, 'r')
	logging.info("Loading ME realignmets...")
	for line in AlignmentFile:
		seq = line.strip('\n').split('\t')
		if seq[0][0] == '@':
			continue
		local_info = parse_name_tp(seq[0])
		Flag = int(seq[1])
		sub_type = seq[2]
		MAPQ = int(seq[4])
		cigar = seq[5]
		cigar_flag = clip_analysis(cigar, args.clipping_threshold)
		if flag_dic[Flag] != 0 and MAPQ >= args.min_mapq and cigar_flag == 1:
			key = "%s*%s*%s*%s"%(local_info.Chr, local_info.Pos, local_info.Len, \
				local_info.GT)
			if key not in cluster_dic:
				cluster_dic[key] = list()
			cluster_dic[key].append("<%s:ME:%s>"%(local_info.Type, sub_type))
	AlignmentFile.close()

	if args.MEI == 'False':
		AlignmentFile = open(path, 'r')
		for line in AlignmentFile:
			seq = line.strip('\n').split('\t')
			if seq[0][0] == '@':
				continue
			local_info = parse_name_tp(seq[0])
			Flag = int(seq[1])
			sub_type = seq[2]
			if sub_type == '*':
				key = "%s*%s*%s*%s"%(local_info.Chr, local_info.Pos, local_info.Len, \
					local_info.GT)
				if key not in cluster_dic:
					cluster_dic[key] = list()
					cluster_dic[key].append("<%s>"%(local_info.Type))
		AlignmentFile.close()

	sort_list = list()
	for i in cluster_dic:
		chr, breakpoint, insert_size, GT = parse_name(i)
		final_type = acquire_count_max(cluster_dic[i])
		sort_list.append([chr, breakpoint, insert_size, final_type])
	sort_list = sorted(sort_list, key = lambda x:(x[0], int(x[1])))
	file = open(out_path, 'w')
	logging.info("Writing results into disk...")
	file.write("# Chromsome\tBreakpoint\tSV length\tMEI Type")
	for i in sort_list:
		file.write("\t".join(i)+"\n")
	file.close()
# ************************BED_FUNCTION*******************************
# 
# 
# 
# ************************VCF_FUNCTION*******************************
def call_vcf(args):
	path = args.input
	out_path = args.output + "calling.vcf"
	ref = load_ref(args.Reference)
	AlignmentFile = open(path, 'r')
	logging.info("Loading ME realignmets...")
	for line in AlignmentFile:
		seq = line.strip('\n').split('\t')
		if seq[0][0] == '@':
			continue
		local_info = parse_seq_head(seq[0])
		Flag = int(seq[1])
		sub_type = seq[2]
		MAPQ = int(seq[4])
		cigar = seq[5]
		cigar_flag = clip_analysis(cigar, args.clipping_threshold)
		if flag_dic[Flag] != 0 and MAPQ >= args.min_mapq and cigar_flag == 1:
			key = "%s*%s*%s*%s"%(local_info.Chr, local_info.Pos, local_info.Len, \
				local_info.GT)
			if key not in cluster_dic:
				cluster_dic[key] = list()
			cluster_dic[key].append("<%s:ME:%s>\t%d"%(local_info.Type, sub_type, \
				flag_dic[Flag]))
	AlignmentFile.close()

	if args.MEI == 'False':
		AlignmentFile = open(path, 'r')
		for line in AlignmentFile:
			seq = line.strip('\n').split('\t')
			if seq[0][0] == '@':
				continue
			local_info = parse_name_tp(seq[0])
			Flag = int(seq[1])
			sub_type = seq[2]
			if sub_type == '*':
				key = "%s*%s*%s*%s"%(local_info.Chr, local_info.Pos, local_info.Len, \
					local_info.GT)
				if key not in cluster_dic:
					cluster_dic[key] = list()
					cluster_dic[key].append("<%s>\t%s"%(local_info.Type, '*'))
		AlignmentFile.close()

	sort_list = list()
	for i in cluster_dic:
		chr, breakpoint, insert_size, GT = parse_name(i)
		final_type = acquire_count_max(cluster_dic[i]).split('\t')[0]
		strand = STRAND[acquire_count_max(cluster_dic[i]).split('\t')[1]]
		sort_list.append([chr, breakpoint, insert_size, final_type, GT, strand])
	sort_list = sorted(sort_list, key = lambda x:(x[0], int(x[1])))
	head_info = print_vcf_head(ref, args.sample)

	file = open(out_path, 'w')
	logging.info("Writing results into disk...")

	for line in head_info:
		file.write(line)

	ID = 0
	for i in sort_list:
		concordant = int(i[4].split(':')[0])
		discordant = int(i[4].split(':')[1]) - int(i[4].split(':')[0])
		if discordant < 0:
			discordant = 0
		GT, GL, reliability = simple_call_genotype(concordant, concordant+discordant, \
			args.heterozygous, args.homozygous)

		if reliability == 1:
			INFO = "PRECISE;SVTYPE=%s;SVLEN=%d;END=%d;SAMPLE=%s;STRAND=%s"%(i[3][1:4], \
				int(i[2]), int(i[1])+int(i[2])-1, args.sample, i[5])
		else:
			INFO = "IMPRECISE;SVTYPE=%s;SVLEN=%d;END=%d;SAMPLE=%s;STRAND=%s"%(i[3][1:4], \
				int(i[2]), int(i[1])+int(i[2])-1, args.sample, i[5])
		try:
			REF = ref[i[0]][int(i[1])-1]
		except:
			REF = "N"
		file.write("%s\t%s\t%d\t%s\t%s\t.\tPASS\t%s\tGT:DV:DR\t%s:%s\n"%(i[0], i[1], \
			ID, REF, i[3], INFO, GT, GL))
		ID += 1
	file.close()
# *************************VCF_FUNCTION*******************************
# 
# 
# 
# ************************MAIN_FUNCTION*******************************
def parseArgs(argv):
	parser = argparse.ArgumentParser(prog="rMETL.py calling", description=USAGE, \
		formatter_class=argparse.RawDescriptionHelpFormatter)
	parser.add_argument("input", metavar="SAM", type=str, help="Input cluster.sam on STAGE realignment.")
	parser.add_argument("Reference", metavar="REFERENCE", type=str, \
		help="The reference genome in fasta format.")
	parser.add_argument("format", metavar="[BED,VCF]", type=str, \
		help="The format of the output file. [%(default)s]", default = "bed")
	parser.add_argument('output', type=str, help = "Directory to output final callset.")
	parser.add_argument('-hom', '--homozygous', \
		help = "The mininum score of a genotyping reported as a homozygous.[%(default)s]", \
		default = 0.8, type = float)
	parser.add_argument('-het','--heterozygous', \
		help = "The mininum score of a genotyping reported as a heterozygous.[%(default)s]", \
		default = 0.3, type = float)
	parser.add_argument('-q', '--min_mapq', help = "Mininum mapping quality.[%(default)s]", \
		default = 20, type = int)
	parser.add_argument('-c', '--clipping_threshold', \
		help = "Mininum threshold of realignment clipping.[%(default)s]", \
		default = 0.5, type = float)
	parser.add_argument('--sample', help = "Sample description", \
		default = "None", type = str)
	parser.add_argument('--MEI', help = "Enables rMETL to display MEI/MED only.[%(default)s]", \
		default = "True", type = str)
	args = parser.parse_args(argv)
	return args

def run(argv):
    args = parseArgs(argv)
    setupLogging(False)
    starttime = time.time()
    if args.format == "bed":
    	call_bed(args)
    elif args.format == "vcf":
    	call_vcf(args)
    else:
    	logging.error("Invalid format.")
    	exit(1)
    logging.info("Finished in %0.2f seconds."%(time.time() - starttime))

if __name__ == '__main__':
	run(sys.argv[:1])



#!/usr/bin/env python
# -*-coding:utf-8-*-

import logging
import os
from Bio import SeqIO
from rMETL.rMETL_cmdRunner import exe

def load_ref(ref_g):
	logging.info("Loading reference genome...")
	return SeqIO.to_dict(SeqIO.parse(ref_g, "fasta"))

def check_bai(file, tempdir):
	'''
	check the index of a BAM file.
	'''
	if os.path.exists(file+".bai"):
		logging.info("The bam file is legal.")
		return ""
	else:
		logging.info("The bam.bai is missed.")
		logging.info("Running Samtools sort...")
		bam_path = file[:-3] + "sorted.bam"
		cmd = ("samtools sort -@ 4 -O bam -T %s -o %s %s" % (tempdir, \
			bam_path, file))
		r, o, e = exe(cmd)
		if r != 0:
			logging.error("Samtools sort failed!")
			logging.error("RETCODE %d" % (r))
			logging.error("STDOUT %s" % (str(o)))
			logging.error("STDERR %s" % (str(e)))
			logging.error("Exiting")
			exit(r)
		logging.info("Finished Samtools sort.")

		logging.info("Running Samtools index...")
		cmd = ("samtools index %s" % (bam_path))
		r, o, e = exe(cmd)
		if r != 0:
			logging.error("Samtools index failed!")
			logging.error("RETCODE %d" % (r))
			logging.error("STDOUT %s" % (str(o)))
			logging.error("STDERR %s" % (str(e)))
			logging.error("Exiting")
			exit(r)
		logging.info("Finished Samtools index.")
		return bam_path

def call_ngmlr(inFile, ref, presets, nproc, outFile):
	"""
	run ngmlr to generate alignments
	"""
	outFile = outFile + "map.sam"
	logging.info("Running NGMLR...")
	cmd = ("ngmlr -r %s -q %s -o %s -t %d -x %s" % (ref, inFile, outFile, \
		nproc, presets))
	r, o, e = exe(cmd)
	
	if r != 0:
		logging.error("NGMLR mapping failed!")
		logging.error("RETCODE %d" % (r))
		logging.error("STDOUT %s" % (str(o)))
		logging.error("STDERR %s" % (str(e)))
		logging.error("Exiting")
		exit(r)
	logging.info("Finished NGMLR mapping.")
	return outFile

def call_samtools(file, tempdir):
	'''
	run samtools to generate sorted BAM files.
	'''
	logging.info("Running Samtools sort...")
	bam_path = file[:-3] + "bam"
	cmd = ("samtools view -Sb %s | samtools sort -@ 4 -O bam -T %s - > %s" % \
		(file, tempdir, bam_path))
	r, o, e = exe(cmd)
	if r != 0:
		logging.error("Samtools sort failed!")
		logging.error("RETCODE %d" % (r))
		logging.error("STDOUT %s" % (str(o)))
		logging.error("STDERR %s" % (str(e)))
		logging.error("Exiting")
		exit(r)
	logging.info("Finished Samtools sort.")

	logging.info("Running Samtools index...")
	cmd = ("samtools index %s" % (bam_path))
	r, o, e = exe(cmd)
	if r != 0:
		logging.error("Samtools index failed!")
		logging.error("RETCODE %d" % (r))
		logging.error("STDOUT %s" % (str(o)))
		logging.error("STDERR %s" % (str(e)))
		logging.error("Exiting")
		exit(r)
	logging.info("Finished Samtools index.")
	return bam_path



# * @author: Jiang Tao (tjiang@hit.edu.cn)

GL_TAG = ['1/0', '0/1', '1/1']

def simple_call_genotype(Nalt, Ntotal, P_heterozygous, P_homozygous):
	bound_low = Ntotal * P_heterozygous
	bound_up = Ntotal * P_homozygous
	if Nalt < bound_low:
        # reliability = 0
		return GL_TAG[0], "%d:%d"%(Nalt, Ntotal - Nalt), 0
	elif bound_low <= Nalt and Nalt < bound_up:
        # reliability = 1
		return GL_TAG[1], "%d:%d"%(Nalt, Ntotal - Nalt), 1
	else:
        # reliability = 1
		return GL_TAG[2], "%d:%d"%(Nalt, Ntotal - Nalt), 1

def simple_filter_genotype(Nalt, Ntotal, P_heterozygous):
    bound_low = Ntotal * P_heterozygous
    if Nalt < bound_low:
        return 0
    else:
        return 1

def count_coverage(chr, s, e, f):
    total = 0
    for i in f.fetch(chr, s, e):
        total += 1
    return total

def add_genotype(info_list, file, low_bandary):
    '''
    allocate genotype for each MEI/MED
    '''
    for i in xrange(len(info_list)):
        if info_list[i][0][0] == 'INS':
            chr = info_list[i][0][1]
            start = info_list[i][0][2]-low_bandary
            end = info_list[i][0][2] + low_bandary
            locus_cov = count_coverage(chr, start, end, file)
            for j in xrange(len(info_list[i])):
                info_list[i][j].append(locus_cov)
        else:
            for j in xrange(len(info_list[i])):
                chr = info_list[i][j][1]
                start = info_list[i][j][2]
                end = info_list[i][j][2]+info_list[i][j][3]
                locus_cov = count_coverage(chr, start, end, file)
                info_list[i][j].append(locus_cov)
    return info_list

if __name__ == '__main__':
	pass



#!/usr/bin/env python

''' 
 * All rights Reserved, Designed By HIT-Bioinformatics   
 * @Title:  extract_simple.py
 * @Package: argparse, pysam, sys, Bio, os, logging
 * @Description: Parse the ME signatures from alignments
 * @author: Jiang Tao (tjiang@hit.edu.cn)
 * @date: Apr 24 2018
 * @version V1.0.2     
'''

import pysam
import cigar
import os
import argparse
import logging
import sys
import time
import gc

from multiprocessing import Pool
from rMETL.rMETL_version import __version__, __author__, __contact__
from rMETL.rMETL_concensus import construct_concensus_info
from rMETL.rMETL_genotype import add_genotype
from rMETL.rMETL_utils import load_ref, check_bai, call_ngmlr, call_samtools
from rMETL.rMETL_cmdRunner import setupLogging, exe

USAGE="""\
           _  ___  _   _____   _______   _
     _ _  | ^_   _^ | |  ___| |__   __| | |
    | ^_| | | | | | | | |__      | |    | |
    | |   | | | | | | |  __|     | |    | |
    | |   | | | | | | | |___     | |    | |___
    |_|   |_| |_| |_| |_____|    |_|    |_____|

    rMETL - realignment-based Mobile Element insertion detection Tool for Long read


	Support reads aligned with Ngmlr and sorted with Samtools

	If input is a fastq or fasta format file, rMETL generates
	alignments with Ngmlr at first;

	If input is a sam format file, rMETL converts and sorts it
	to be a bam format file;

	If your input is a bam format file with index, rMETL extracts
	the ME signatures and collects the sub-sequence of them.

	The output is a fasta format file called 'potential.fa' 
	contains potentials non-reference ME clusters.

	rMETL V%s
	Author: %s
	Contact: %s
"""%(__version__, __author__, __contact__)

INS_flag = {1:'I'}
DEL_flag = {2:'D'}
clip_flag = {4:'S', 5:'H'}
global_ref = list()

# **********************check-input-format****************************
def decipherInput(input):
	"""
	resolution input format
	"""
	extension = input.split('.')[-1].lower()
	choice = {"bam": 0, \
			  "sam": 1, \
			  "fasta": 2, \
			  "fastq": 2, \
			  "fa": 2, \
			  "fq": 2}
	return choice[extension]
# **********************check-input-format****************************
# 
# 
# 
# ************************mini-operations*****************************
def revcom_complement(s): 
	'''
	generation reverse complementary sequence.
	all of the lowercase will be changed to capital letter.
	'''
	basecomplement = {'A': 'T', \
					  'C': 'G', \
					  'G': 'C', \
					  'T': 'A', \
					  'a': 'T', \
					  'c': 'G', \
					  'g': 'C', \
					  't': 'A'} 
	letters = list(s) 
	letters = [basecomplement[base] for base in letters] 
	return ''.join(letters)[::-1]

def detect_flag(Flag):
	'''
	identification of flag in BAM
		0 means unmapped read
		1 & 2 means primary mapping read with normal strand or reverse strand 
		3 & 4 means supplementary mapping read with normal strand or reverse strand
	'''
	Normal_foward = 1 >> 1
	Abnormal = 1 << 2
	Reverse_complement = 1 << 4
	Supplementary_map = 1 << 11

	signal = {Abnormal: 0, \
			  Normal_foward: 1, \
			  Reverse_complement: 2, \
			  Supplementary_map: 3, \
			  Reverse_complement | Supplementary_map: 4}

	back_sig = signal[Flag] if Flag in signal else 0
	return back_sig

def acquire_clip_pos(deal_cigar):
	'''
	resolution of cigar in supplementary mapping
	'''
	seq = list(cigar.Cigar(deal_cigar).items())
	first_pos = seq[0][0] if seq[0][1] == 'S' else 0
	last_pos = seq[-1][0] if seq[-1][1] == 'S' else 0
	bias = 0
	for i in seq:
		if i[1] in ['M', 'D']:
			bias += i[0]
	return [first_pos, last_pos, bias]
# ************************mini-operations*****************************
# 
# 
# 
# ************************soft-clippings******************************
def store_clip_pos(locus, chr, seq, flag, CLIP_note):
	'''
	A data structure store info of soft-clippings.
	It has two hashtables:
		1. key1 is an integer calculated by coordinate / 10000;
		2. key2 is an integer calculated by (coordinate % 10000) / 50.
	'''
	hash_1 = int(locus /10000)
	mod = locus % 10000
	hash_2 = int(mod / 50)
	element = [locus, seq, flag]
	if hash_1 not in CLIP_note[chr]:
		CLIP_note[chr][hash_1] = dict()
		CLIP_note[chr][hash_1][hash_2] = list()
		CLIP_note[chr][hash_1][hash_2].append(element)
	else:
		if hash_2 not in CLIP_note[chr][hash_1]:
			CLIP_note[chr][hash_1][hash_2] = list()
			CLIP_note[chr][hash_1][hash_2].append(element)
		else:
			CLIP_note[chr][hash_1][hash_2].append(element)

def acquire_clip_locus(down, up, chr, CLIP_note):
	'''
	search soft-clippings within limited region
	'''
	list_clip = list()
	if int(up/10000) == int(down/10000):
		key_1 = int(down/10000)
		if key_1 not in CLIP_note[chr]:
			return list_clip
		for i in xrange(int((up%10000)/50)-int((down%10000)/50)+1):
			key_2 = int((down%10000)/50)+i
			if key_2 not in CLIP_note[chr][key_1]:
				continue
			for ele in CLIP_note[chr][key_1][key_2]:
				if ele[0] >= down and ele[0] <= up:
					list_clip.append(ele)
	else:
		key_1 = int(down/10000)
		if key_1 in CLIP_note[chr]:
			for i in xrange(200-int((down%10000)/50)):
				key_2 = int((down%10000)/50)+i
				if key_2 not in CLIP_note[chr][key_1]:
					continue
				for ele in CLIP_note[chr][key_1][key_2]:
					if ele[0] >= down and ele[0] <= up:
						list_clip.append(ele)
		key_1 += 1
		if key_1 not in CLIP_note[chr]:
			return list_clip
		for i in xrange(int((up%10000)/50)+1):
			key_2 = i
			if key_2 not in CLIP_note[chr][key_1]:
				continue
			for ele in CLIP_note[chr][key_1][key_2]:
				if ele[0] >= down and ele[0] <= up:
					list_clip.append(ele)
	return list_clip
# ************************soft-clippings******************************
# 
# 
# 
# ***********************resolution-reads*****************************
def organize_split_signal(chr, primary_info, Supplementary_info, \
	total_L, low_bandary):
	'''
	resolution split alignments
	'''
	overlap = list()
	for i in Supplementary_info:
		seq = i.split(',')
		local_chr = seq[0]
		local_start = int(seq[1])
		local_cigar = seq[3]
		dic_starnd = {1:'+', 2: '-'}
		if dic_starnd[primary_info[4]] != seq[2]:
			continue
		if chr != local_chr:
			continue
		local_set = acquire_clip_pos(local_cigar)
		if primary_info[0] < local_start:
			if primary_info[3]+local_set[0]-total_L > low_bandary:
				overlap.append([total_L - primary_info[3], \
					local_set[0], primary_info[1]])
		else:
			if local_set[1]+primary_info[2]-total_L > low_bandary:
				overlap.append([total_L - local_set[1], \
					primary_info[2], local_start+local_set[2]-1])
	return overlap

def parse_read(read, Chr_name, low_bandary, CLIP_note):
	'''
	Check:	1.Flag
			2.Supplementary mapping
			3.Seq
	'''
	DEL_ME_pos = list()
	INS_ME_pos = list()
	process_signal = detect_flag(read.flag)
	if process_signal == 0:
		return INS_ME_pos, DEL_ME_pos

	# Add DEL:ME type call signal
	pos_start = read.reference_start
	shift = 0
	for element in read.cigar:
		if element[0] == 0:
			shift += element[1]
		if element[0] in DEL_flag and element[1] <= low_bandary:
			shift += element[1]
		if element[0] in DEL_flag and element[1] > low_bandary:
			DEL_ME_pos.append([pos_start+shift, element[1]])
			shift += element[1]

	# Add INS:ME type call signal
	pos_start = read.reference_start
	shift = 0
	_shift_read_ = 0
	pos_end = read.reference_end
	primary_clip_0 = 0
	primary_clip_1 = 0
	for element in read.cigar:
		if element[0] == 0 or element[0] == 2:
			shift += element[1]
		if element[0] != 2:
			_shift_read_ += element[1]
		if element[0] in INS_flag and element[1] > low_bandary:
			shift += 1
			MEI_contig = read.query_sequence[_shift_read_ - \
			element[1]:_shift_read_]
			INS_ME_pos.append([pos_start + shift, element[1], \
				MEI_contig])
		if element[0] in clip_flag:
			if shift == 0:
				primary_clip_0 = element[1]
			else:
				primary_clip_1 = element[1]
			if element[1] > low_bandary:
				if shift == 0:
					clip_pos = pos_start - 1
					clip_contig = read.query_sequence[:element[1]]
					store_clip_pos(clip_pos, Chr_name, clip_contig, \
						0, CLIP_note)
				else:
					clip_pos = pos_start + shift - 1
					clip_contig = read.query_sequence[read.query_length \
					- element[1]:]
					store_clip_pos(clip_pos, Chr_name, clip_contig, 1, \
						CLIP_note)

	if process_signal in [1, 2]:
		Tags = read.get_tags()
		chr = Chr_name
		primary_info = [pos_start, pos_end, primary_clip_0, primary_clip_1, \
		process_signal]
		for i in Tags:
			if i[0] == 'SA':
				Supplementary_info = i[1].split(';')[:-1]
				overlap = organize_split_signal(chr, primary_info, \
					Supplementary_info, read.query_length, low_bandary)
				for k in overlap:
					MEI_contig = read.query_sequence[k[0]:k[1]]
					INS_ME_pos.append([k[2], k[1] - k[0], MEI_contig])
	return INS_ME_pos, DEL_ME_pos
# ***********************resolution-reads*****************************
# 
# 
# 
# ***********************Cluster-Function*****************************
def merge_pos(pos_list, chr, evidence_read, SV_size, CLIP_note):
	'''
	INS: inner-cluster function
	'''
	start = list()
	end = list()
	for ele in pos_list:
		start.append(ele[0])
		end.append(ele[0] + ele[1])
	search_down = min(start) - 10
	search_up = max(start) + 10
	temp_clip = acquire_clip_locus(search_down, search_up, chr, CLIP_note)
	result = construct_concensus_info(pos_list, temp_clip, evidence_read, \
		SV_size)
	if result != 0:
		for i in xrange(len(result)):
			result[i] = ["INS", chr] + result[i] + [len(result)]
		return result
	else:
		return 0

def cluster(pos_list, chr, evidence_read, SV_size, low_bandary, CLIP_note):
	'''
	INS: outer-cluster function
	'''
	_cluster_ = list()
	temp = list()
	temp.append(pos_list[0])
	for pos in pos_list[1:]:
		if temp[-1][0] + low_bandary < pos[0]:
			result = merge_pos(temp, chr, evidence_read, SV_size, CLIP_note)
			if result != 0:
				_cluster_.append(result)
			temp = list()
			temp.append(pos)
		else:
			temp.append(pos)
	result = merge_pos(temp, chr, evidence_read, SV_size, CLIP_note)
	if result != 0:
		_cluster_.append(result)
	return _cluster_

def merge_pos_del(pos_list, chr, Ref, evidence_read, SV_size):
	'''
	DEL: inner-cluster function
	'''
	start = list()
	end = list()
	for ele in pos_list:
		start.append(ele[0])
		end.append(ele[0] + ele[1])
	breakpoint = sum(start)/len(start)
	size = sum(end)/len(end) - breakpoint
	result = list()
	if len(pos_list) < evidence_read:
		return result
	else:
		if chr in Ref and size >= SV_size:
			result.append(['DEL', chr, breakpoint, size, len(pos_list), \
				str(Ref[chr].seq[breakpoint:breakpoint+size])])
	return result

def cluster_del(pos_list, chr, Ref, evidence_read, SV_size, low_bandary):
	'''
	DEL: outer-cluster function
	'''
	_cluster_ = list()
	temp = list()
	temp.append(pos_list[0])
	for pos in pos_list[1:]:
		if temp[-1][0] + low_bandary < pos[0]:
			result = merge_pos_del(temp, chr, Ref, evidence_read, SV_size)
			if len(result) != 0:
				_cluster_.append(result)
			temp = list()
			temp.append(pos)
		else:
			temp.append(pos)
	result = merge_pos_del(temp, chr, Ref, evidence_read, SV_size)
	if len(result) != 0:
		_cluster_.append(result)
	return _cluster_
# ***********************Cluster-Function*****************************
# 
# 
# 
# ***********************Output-Function******************************
def combine_result(INS, DEL, path, chr):
	'''
	Merge results into one list and output it.
	'''
	output = "%ssignatures/%s_sig.fa"%(path, chr)
	file = open(output, 'w')
	for i in INS:
		for j in i:
			if len(j) != 8:
				continue
			key = "%s*%s*%d*%d*%s*%d*%d"%(j[0], j[1], j[2], j[3], j[4], j[6], \
				j[7])
			file.write(">"+key+'\n')
			file.write(j[5]+'\n')
	del INS
	gc.collect()
	for i in DEL:
		for j in i:
			if len(j) != 7:
				continue
			key = "%s*%s*%d*%d*%d*%d"%(j[0], j[1], j[2], j[3], j[4], j[6])
			file.write(">%s\n"%(key))
			file.write('%s\n'%(j[5]))
	del DEL
	gc.collect()
	file.close()
# ***********************Output-Function******************************
# 
# 
# 
# ********************Signatures-extraction***************************
def single_pipe(out_path, chr, bam_path, low_bandary, evidence_read, SV_size):
	'''
	resolution signatures
	'''
	samfile = pysam.AlignmentFile(bam_path)
	CLIP_note = dict()
	logging.info("Resolving chromsome %s."%(chr))
	if chr not in CLIP_note:
		CLIP_note[chr] = dict()
	cluster_pos_INS = list()
	cluster_pos_DEL = list()
	for read in samfile.fetch(chr):
		feed_back, feed_back_del = parse_read(read, chr, low_bandary, CLIP_note)
		cluster_pos_INS += feed_back
		cluster_pos_DEL += feed_back_del
	cluster_pos_INS = sorted(cluster_pos_INS, key = lambda x:x[0])
	cluster_pos_DEL = sorted(cluster_pos_DEL, key = lambda x:x[0])
	if len(cluster_pos_INS) == 0:
		Cluster_INS = list()
	else:
		Cluster_INS = cluster(cluster_pos_INS, chr, evidence_read, SV_size, \
			low_bandary, CLIP_note)
		del cluster_pos_INS
		del CLIP_note[chr]
		gc.collect()
	if len(cluster_pos_DEL) == 0:
		Cluster_DEL = list()
	else:
		Ref = global_ref[0]
		Cluster_DEL = cluster_del(cluster_pos_DEL, chr, Ref, evidence_read, \
			SV_size, low_bandary)
		del cluster_pos_DEL
		gc.collect()
	logging.info("%d MEI/MED signal loci in the chromsome %s."%(len(Cluster_INS)+\
		len(Cluster_DEL), chr))
	combine_result(add_genotype(Cluster_INS, samfile, low_bandary), \
		add_genotype(Cluster_DEL, samfile, low_bandary), out_path, chr)
	samfile.close()

def multi_run_wrapper(args):
   return single_pipe(*args)

def load_sam_multi_processes(args):
	'''
	task scheduling
	'''
	temporary_dir = args.temp_dir if args.temp_dir.endswith('/') else \
	"%s/"%(args.temp_dir)
	os.mkdir("%ssignatures"%temporary_dir)
	# Major Steps:
	# loading alignment file: bam format
	samfile = pysam.AlignmentFile(args.input)
	# loading reference genome
	Ref = load_ref(args.Reference)
	global_ref.append(Ref)
	# acquire the total numbers of the ref contigs
	contig_num = len(samfile.get_index_statistics())
	logging.info("The total number of chromsomes: %d"%(contig_num))
	# Thread scheduling
	process_list = list()
	for i in samfile.get_index_statistics():
		process_list.append([i[0], i[3]])
		# #chr #read
	process_list = sorted(process_list, key = lambda x:-x[1])
	# start to establish multiprocesses
	analysis_pools = Pool(processes = args.threads)
	# Acquire_Chr_name
	for i in process_list:
		para = [(temporary_dir, i[0], args.input, args.min_distance, \
			args.min_support, args.min_length)]
		analysis_pools.map_async(multi_run_wrapper, para)
	analysis_pools.close()
	analysis_pools.join()
	samfile.close()

	output_p = args.output_dir if args.output_dir.endswith('/') else \
	"%s/"%(args.output_dir)
	if not os.path.exists(output_p):
		os.mkdir(output_p)
	merge_cmd = ("cat %ssignatures/* > %spotential_ME.fa"%(temporary_dir, output_p))
	r, o, e = exe(merge_cmd)
	if r != 0:
		logging.error("Merging ME signatures failed!")
		logging.error("RETCODE %d" % (r))
		logging.error("STDOUT %s" % (str(o)))
		logging.error("STDERR %s" % (str(e)))
		logging.error("Exiting")
		exit(r)
	logging.info("Cleaning temporary files.")
	cmd_remove_tempfile = ("rm -r %ssignatures"%(temporary_dir))
	r, o, e = exe(cmd_remove_tempfile)
	if r != 0:
		logging.error("Cleaning temporary files failed!")
		logging.error("RETCODE %d" % (r))
		logging.error("STDOUT %s" % (str(o)))
		logging.error("STDERR %s" % (str(e)))
		logging.error("Exiting")
		exit(r)
# ********************Signatures-extraction***************************
# 
# 
# 
# ************************MAIN_FUNCTION*******************************
def parseArgs(argv):
	parser = argparse.ArgumentParser(prog="rMETL.py detection", \
		description=USAGE, formatter_class=argparse.RawDescriptionHelpFormatter)
	parser.add_argument("input", metavar="[SAM,BAM,FASTA,FASTQ]", type=str, \
		help="Input reads with/without alignment.")
	parser.add_argument("Reference", metavar="REFERENCE", type=str, \
		help="The reference genome in fasta format.")
	parser.add_argument('temp_dir', type=str, \
		help = "Temporary directory to use for distributed jobs.")
	parser.add_argument('output_dir', type=str, \
		help = "Directory to output potential ME loci.")
	parser.add_argument('-s', '--min_support',\
	 help = "Mininum number of reads that support a ME.[%(default)s]", \
	 default = 5, type = int)
	parser.add_argument('-l', '--min_length', \
		help = "Mininum length of ME to be reported.[%(default)s]", \
		default = 50, type = int)
	parser.add_argument('-d', '--min_distance', \
		help = "Mininum distance of two ME signatures to be intergrated.[%(default)s]", \
		default = 20, type = int)
	parser.add_argument('-t', '--threads', \
		help = "Number of threads to use.[%(default)s]", default = 8, \
		type = int)
	parser.add_argument('-x', '--presets', \
		help = "The sequencing platform <pacbio,ont> of the reads.[%(default)s]", \
		default = "pacbio", type = str)
	args = parser.parse_args(argv)
	return args

def run(argv):
	args = parseArgs(argv)
	setupLogging(False)
	starttime = time.time()
	flag = decipherInput(args.input)
	
	if flag == 0:
		# detection
		result = check_bai(args.input, args.temp_dir)
		if len(result) == 0:
			load_sam_multi_processes(args)
		else:
			args.input = result
			load_sam_multi_processes(args)
	elif flag == 1:
		bam_path = call_samtools(args.input, args.temp_dir)
		args.input = bam_path
		load_sam_multi_processes(args)
		# samtools transfer
	else:
		# inFile, ref, seq_type, nproc=1, outFile="map.sam", presets="pacbio"
		file = call_ngmlr(args.input, args.Reference, args.presets, \
			rgs.threads, args.temp_dir)
		bam_path = call_samtools(file, args.temp_dir)
		args.input = bam_path
		load_sam_multi_processes(args)
	logging.info("Finished in %0.2f seconds."%(time.time() - starttime))

if __name__ == '__main__':
	run(sys.argv[:1])



# coding=utf-8

from setuptools import setup, find_packages

LONG_DESCRIPTION = '''Mobile element insertion (MEI) is a major category of structure variations (SVs). \
The rapid development of long read sequencing provides the opportunity to sensitively discover MEIs. \
However, the signals of MEIs implied by noisy long reads are highly complex, due to the repetitiveness \
of mobile elements as well as the serious sequencing errors. Herein, we propose Realignment-based \
Mobile Element insertion detection Tool for Long read (rMETL). rMETL takes advantage of \
its novel chimeric read re-alignment approach to well handle complex MEI signals. \
Benchmarking results on simulated and real datasets demonstrated that rMETL has the ability \
to more sensitivity discover MEIs as well as prevent false positives. \
It is suited to produce high quality MEI callsets in many genomics studies.'''

setup(
    name = "rMETL",
    version = "1.0.4",
    description = "realignment-based Mobile Element insertion detection Tool for Long read",
    author = "Jiang Tao",
    author_email = "tjiang@hit.edu.cn",
    url = "https://github.com/tjiangHIT/rMETL",
    license = "MIT",
    packages = find_packages("src"),
    package_dir = {"": "src"},
    data_files = [("", ["LICENSE"])],
    scripts=['src/rMETL/rMETL'],
    long_description = LONG_DESCRIPTION,
    zip_safe = False,
    install_requires = ['pysam', 'Biopython', 'Cigar']
)



#!/usr/bin/env python

''' 
 * All rights Reserved, Designed By HIT-Bioinformatics   
 * @Title:  Map.py
 * @Package: argparse, sys, logging
 * @Description: Classify the ME types
  * @author: Jiang Tao (tjiang@hit.edu.cn)
 * @date: Apr 24 2018
 * @version V1.0.2
'''

import argparse
import logging
import sys
import time

from rMETL.rMETL_version import __version__, __author__, __contact__
from rMETL.rMETL_cmdRunner import setupLogging, exe

USAGE="""\
           _  ___  _   _____   _______   _
     _ _  | ^_   _^ | |  ___| |__   __| | |
    | ^_| | | | | | | | |__      | |    | |
    | |   | | | | | | |  __|     | |    | |
    | |   | | | | | | | |___     | |    | |___
    |_|   |_| |_| |_| |_____|    |_|    |_____|

    rMETL - realignment-based Mobile Element insertion detection Tool for Long read

	Realignment of chimeric read parts.

	Aligner: NGMLR version 0.2.6
	TE refs: Alu concensus
		 L1 concensus
		 SVA concensus
	The output is a sam format file called 'cluster.sam'.

	rMETL V%s
	Author: %s
	Contact: %s
"""%(__version__, __author__, __contact__)

# **************************Call-NGMLR********************************
def call_ngmlr(inFile, ref, presets, nproc, outFile, SUBREAD_LENGTH, SUBREAD_CORRIDOR):
	"""
	fq = input file
	automatically search for .sa
	"""
	outFile = outFile + "cluster.sam"
	logging.info("Running NGMLR...")
	cmd = ("ngmlr -r %s -q %s -o %s -t %d -x %s --subread-length %d --subread-corridor %d" \
		% (ref, inFile, outFile, nproc, presets, SUBREAD_LENGTH, SUBREAD_CORRIDOR))
	r, o, e = exe(cmd)
	if r != 0:
		logging.error("NGMLR mapping failed!")
		logging.error("RETCODE %d" % (r))
		logging.error("STDOUT %s" % (str(o)))
		logging.error("STDERR %s" % (str(e)))
		logging.error("Exiting")
		exit(r)
	logging.info("Finished NGMLR mapping.")
	return outFile
# **************************Call-NGMLR********************************
# 
# 
# 
# ************************MAIN_FUNCTION*******************************
def parseArgs(argv):
	parser = argparse.ArgumentParser(prog="rMETL.py realignment", description=USAGE, \
		formatter_class=argparse.RawDescriptionHelpFormatter)
	parser.add_argument("input", metavar="FASTA", type=str, help="Input potential_ME.fa on STAGE detection.")
	parser.add_argument("ME_Ref", type=str, help="The transposable element concensus in fasta format.")
	parser.add_argument('output', type=str, help = "Directory to output realignments.")
	parser.add_argument('-t', '--threads', help = "Number of threads to use.[%(default)s]", \
		default = 8, type = int)
	parser.add_argument('-x', '--presets', \
		help = "The sequencing platform <pacbio,ont> of the reads.[%(default)s]", \
		default = "pacbio", type = str)
	parser.add_argument('--subread_length', \
		help = "Length of fragments reads are split into [%(default)s]", \
		default = 128, type = int)
	parser.add_argument('--subread_corridor', \
		help = "Length of corridor sub-reads are aligned with [%(default)s]", \
		default = 20, type = int)
	args = parser.parse_args(argv)
	return args

def run(argv):
	args = parseArgs(argv)
	setupLogging(False)
	starttime = time.time()
	call_ngmlr(args.input, args.ME_Ref, args.presets, args.threads, args.output, \
		args.subread_length, args.subread_corridor)
	logging.info("Finished in %0.2f seconds."%(time.time() - starttime))

if __name__ == '__main__':
    run(sys.argv[:1])
